/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.descriptor.repo;

import com.fasterxml.jackson.annotation.JsonProperty;
import org.jfrog.common.config.diff.GenerateDiffFunction;

import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import java.util.Comparator;

/**
 * @author Shay Yaakov
 */
@XmlEnum
@GenerateDiffFunction(typeMethod = "getType")
public enum RepoType {
    @XmlEnumValue("maven") @JsonProperty("Maven") Maven("maven", "Maven", "gav"),
    @XmlEnumValue("gradle") @JsonProperty("Gradle") Gradle("gradle", "Gradle", "gav"),
    @XmlEnumValue("ivy") @JsonProperty("Ivy") Ivy("ivy", "Ivy"),
    @XmlEnumValue("sbt") @JsonProperty("SBT") SBT("sbt", "SBT", "gav"),
    @XmlEnumValue("nuget") @JsonProperty("NuGet") NuGet("nuget", "NuGet"),
    @XmlEnumValue("gems") @JsonProperty("Gems") Gems("gems", "Gems", "rubygems"),
    @XmlEnumValue("npm") @JsonProperty("Npm") Npm("npm", "npm"),
    @XmlEnumValue("bower") @JsonProperty("Bower") Bower("bower", "Bower"),
    @XmlEnumValue("debian") @JsonProperty("Debian") Debian("debian", "Debian", "deb"),
    @XmlEnumValue("pypi") @JsonProperty("Pypi") Pypi("pypi", "Pypi"),
    @XmlEnumValue("cargo") @JsonProperty("Cargo") Cargo("cargo", "Cargo"),
    @XmlEnumValue("puppet") @JsonProperty("Puppet") Puppet("puppet", "Puppet"),
    @XmlEnumValue("docker") @JsonProperty("Docker") Docker("docker", "Docker"),
    @XmlEnumValue("vagrant") @JsonProperty("Vagrant") Vagrant("vagrant", "Vagrant"),
    @XmlEnumValue("gitlfs") @JsonProperty("GitLfs") GitLfs("gitlfs", "GitLfs"),
    @XmlEnumValue("yum") @JsonProperty("YUM") YUM("yum", "RPM", "rpm"),
    @XmlEnumValue("vcs") @JsonProperty("VCS") VCS("vcs", "VCS"),
    @XmlEnumValue("p2") @JsonProperty("P2") P2("p2", "P2"),
    @XmlEnumValue("generic") @JsonProperty("Generic") Generic("generic", "Generic"),
    @XmlEnumValue("opkg") @JsonProperty("Opkg") Opkg("opkg", "Opkg"),
    @XmlEnumValue("cocoapods") @JsonProperty("CocoaPods") CocoaPods("cocoapods", "CocoaPods"),
    @XmlEnumValue("conan") @JsonProperty("Conan") Conan("conan", "Conan"),
    @XmlEnumValue("distribution") @JsonProperty("Distribution") Distribution("distribution", "Distribution"),
    @XmlEnumValue("releasebundles") @JsonProperty("ReleaseBundles") ReleaseBundles("releasebundles", "ReleaseBundles"),
    @XmlEnumValue("buildinfo") @JsonProperty("BuildInfo") BuildInfo("buildinfo", "BuildInfo"),
    @XmlEnumValue("pipeinfo") @JsonProperty("PipeInfo") PipeInfo("pipeinfo", "PipeInfo"),
    @XmlEnumValue("chef") @JsonProperty("Chef") Chef("chef", "Chef"),
    @XmlEnumValue("composer") @JsonProperty("Composer") Composer("composer", "Composer"),
    @XmlEnumValue("helm") @JsonProperty("Helm") Helm("helm", "Helm"),
    @XmlEnumValue("go") @JsonProperty("Go") Go("go", "Go"),
    @XmlEnumValue("cran") @JsonProperty("CRAN") CRAN("cran", "CRAN"),
    @XmlEnumValue("conda") @JsonProperty("Conda") Conda("conda", "Conda"),
    @XmlEnumValue("alpine") @JsonProperty("Alpine") Alpine("alpine", "Alpine Linux"),
    @XmlEnumValue("supportbundles") @JsonProperty("Support") Support("support", "Support");

    private final String type;
    private final String displayName;
    private String pkgidPrefix;

    RepoType(String type, String displayName) {
        this.type = type;
        this.displayName = displayName;
        this.pkgidPrefix = type;
    }

    RepoType(String type, String displayName, String pkgidPrefix) {
        this.type = type;
        this.displayName = displayName;
        this.pkgidPrefix = pkgidPrefix;
    }

    public String getType() {
        return type;
    }

    public String getDisplayName() {
        return displayName;
    }

    // Used by REST methods to properly display package types to the users
    public String getEffectiveType() {
        if (this == YUM) {
            return "rpm";
        }
        return type;
    }

    public boolean isMavenGroup() {
        return this == Maven || this == Ivy || this == Gradle || this == P2 || this == SBT;
    }

    public boolean isPipeInfo() {
        return this == PipeInfo;
    }

    public String getPkgidPrefix() {
        if (!pkgidPrefix.equals(type)) {
            return pkgidPrefix;
        }
        return type;
    }

    public boolean isVcsGroup() {
        return this == VCS || this == Composer || this == CocoaPods || this == Bower;
    }

    public static RepoType fromType(String type) {
        type = normalizePackageType(type);
        for (RepoType repoType : values()) {
            if (type.equalsIgnoreCase(repoType.type)) {
                return repoType;
            }
        }
        return Generic;
    }
    // Used to convert users requests (REST \ UI) to backend type

    private static String normalizePackageType(String packageType) {
        packageType = packageType.toLowerCase();
        if (packageType.equals("rpm")) {
            return "yum";
        }
        return packageType;
    }

    public static class RepoNameComparator implements Comparator<RepoType> {

        @Override
        public int compare(RepoType o1, RepoType o2) {
            return o1.name().compareTo(o2.name());
        }
    }
}
