/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.binstore.itest.service;

import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.environment.BinaryStoreProperties;
import org.artifactory.storage.db.binstore.service.BlobBinaryProvider;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.jfrog.storage.binstore.config.BinaryProviderConfigBuilder;
import org.jfrog.storage.binstore.config.model.ChainMetaData;
import org.jfrog.storage.binstore.config.model.Config;
import org.jfrog.storage.binstore.config.model.ProviderMetaData;
import org.jfrog.storage.binstore.ifc.BinaryProviderConfig;
import org.jfrog.storage.binstore.manager.BinaryProviderManagerImpl;
import org.jfrog.storage.binstore.providers.EmptyBinaryProvider;
import org.jfrog.storage.binstore.providers.FileBinaryProvider;
import org.jfrog.storage.binstore.providers.FileBinaryProviderImpl;
import org.jfrog.storage.binstore.providers.FileCacheBinaryProviderImpl;
import org.jfrog.storage.binstore.providers.base.BinaryProviderBase;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.xml.bind.JAXBException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Collections;
import java.util.Optional;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicBoolean;

import static org.jfrog.common.ResourceUtils.getResourceAsFile;
import static org.jfrog.storage.binstore.config.ConfigurableBinaryProviderManager.buildByConfig;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

/**
 * @author Gidi Shabat
 */
@Test
public class ConfigurableBinaryProviderManagerTest extends DbBaseTest {

    private String dataDir;
    private String securityDir;
    private BinaryProviderManagerImpl manager;

    @BeforeClass
    private void init() {
        ArtifactoryHome artifactoryHome = ArtifactoryHome.get();
        dataDir = artifactoryHome.getDataDir().getAbsolutePath();
        securityDir = artifactoryHome.getSecurityDir().getAbsolutePath();
    }

    @AfterClass
    private void deleteDataAndSecurityDirs() {
        new File(dataDir).delete();
        new File(securityDir).delete();
    }

    @Test
    public void binaryProviderWithOverrideProviderTest() {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystoreWithOverideProviders.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        manager = new BinaryProviderManagerImpl(defaultValues);
        BinaryProviderBase blobBinaryProvider = (BinaryProviderBase) ReflectionTestUtils.getField(manager,
                "firstBinaryProvider");
        Assert.assertTrue(blobBinaryProvider instanceof BlobBinaryProvider);
        BinaryProviderBase empty = blobBinaryProvider.next();
        Assert.assertTrue(empty instanceof EmptyBinaryProvider);
    }

    @Test
    public void binaryProviderWithTemplateTest() {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystore-filesystem-template.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        manager = new BinaryProviderManagerImpl(defaultValues);
        BinaryProviderBase fileSystem = (BinaryProviderBase) ReflectionTestUtils.getField(manager,
                "firstBinaryProvider");
        Assert.assertTrue(fileSystem instanceof FileBinaryProviderImpl);
        BinaryProviderBase empty = fileSystem.next();
        Assert.assertTrue(empty instanceof EmptyBinaryProvider);
        Assert.assertEquals("filestore", ((FileBinaryProvider) fileSystem).getBinariesDir().getName());
    }

    @Test
    public void binaryProviderWithExistingProviderTest() {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystoreWithExistingProviders.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        manager = new BinaryProviderManagerImpl(defaultValues);
        BinaryProviderBase blobBinaryProvider = (BinaryProviderBase) ReflectionTestUtils.getField(manager,
                "firstBinaryProvider");
        Assert.assertTrue(blobBinaryProvider instanceof BlobBinaryProvider);
        BinaryProviderBase empty = blobBinaryProvider.next();
        Assert.assertTrue(empty instanceof EmptyBinaryProvider);
    }

    @Test
    public void binaryProviderWithTemplateAndProviderTest() {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystoreWithTemplateAndProvider.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        manager = new BinaryProviderManagerImpl(defaultValues);
        BinaryProviderBase cache = (BinaryProviderBase) ReflectionTestUtils.getField(manager,
                "firstBinaryProvider");
        Assert.assertTrue(cache instanceof FileCacheBinaryProviderImpl);
        BinaryProviderBase fileSystem = cache.next();
        Assert.assertTrue(fileSystem instanceof FileBinaryProviderImpl);
        BinaryProviderBase empty = fileSystem.next();
        Assert.assertTrue(empty instanceof EmptyBinaryProvider);
        Assert.assertEquals("test654", ((FileBinaryProvider) fileSystem).getBinariesDir().getName());
    }

    @Test
    public void binaryProviderWithUserChainTest() {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystoreWithUserChain.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        manager = new BinaryProviderManagerImpl(defaultValues);
        BinaryProviderBase cacheFs = (BinaryProviderBase) ReflectionTestUtils.getField(manager,
                "firstBinaryProvider");
        Assert.assertTrue(cacheFs instanceof FileCacheBinaryProviderImpl);
        BinaryProviderBase retry = cacheFs.next();
        BinaryProviderBase fileSystem2 = retry.next();
        Assert.assertTrue(fileSystem2 instanceof FileBinaryProviderImpl);
        BinaryProviderBase empty = fileSystem2.next();
        Assert.assertTrue(empty instanceof EmptyBinaryProvider);
        Assert.assertEquals("test89", ((FileBinaryProvider) cacheFs).getBinariesDir().getName());
        Assert.assertEquals("test99", ((FileBinaryProvider) fileSystem2).getBinariesDir().getName());
    }

    @Test
    public void signalMirrorChangedTest() {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystoreWithUserChain.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        manager = new BinaryProviderManagerImpl(defaultValues);
        AtomicBoolean wasHit = new AtomicBoolean();
        Assert.assertNull(ReflectionTestUtils.getField(manager,
                "mirrorChangeListener"));
        manager.registerMirrorChangedListener((membersToAdd, membersToDelete, defaultValues1) -> wasHit.set(true));
        Assert.assertNotNull(ReflectionTestUtils.getField(manager,
                "mirrorChangeListener"));
        manager.signalMirrorChanged(Collections.emptyList(), Collections.emptyList());
        Assert.assertTrue(wasHit.get());
    }

    @Test(dataProvider = "addFederationProvider")
    public void creationOfChainTest(boolean shouldAddFederation) throws IOException, JAXBException {
        BinaryProviderConfig defaultValues = new BinaryStoreProperties(dataDir, securityDir).toDefaultValues();
        File configFile = getResourceAsFile("/binarystore/config/binarystoreWithUserChain.xml");
        defaultValues.setBinaryStoreXmlPath(configFile.getAbsolutePath());
        Config config =
                BinaryProviderConfigBuilder.loadConfig(new FileInputStream(defaultValues.getBinaryStoreXmlPath()));
        ChainMetaData chainMetaData = buildByConfig(config, defaultValues, Optional.empty(), shouldAddFederation);
        ProviderMetaData providerMetaData = chainMetaData.getProviderMetaData();
        if (shouldAddFederation) {
            Assert.assertEquals(providerMetaData.getId(), "federated-repo");
            assertEquals(providerMetaData.getType(), "federated-repo",
                    "Expecting providerMetaData type to be federated-repo");
            assertEquals(providerMetaData.getId(), "federated-repo",
                    "Expecting providerMetaData id to be federated-repo");
            assertEquals(providerMetaData.getParams().size(), 5);
            assertEquals(providerMetaData.getParamValue("numberOfRemoteImporters"), "6");
            assertEquals(providerMetaData.getParamValue("numberOfLocalImporters"), "6");
            assertEquals(providerMetaData.getParamValue("maxRetry"), "10");
            assertEquals(providerMetaData.getParamValue("errorRecoveryInterval"), "30000");
            assertEquals(providerMetaData.getParamValue("baseDataDir"), defaultValues.getParam("baseDataDir"));
            assertEquals(providerMetaData.getProviderMetaData().getId(), config.getProviderMetaDatas().get(0).getId());
            assertTrue(providerMetaData.getSubProviderMetaData() instanceof CopyOnWriteArrayList);
        } else {
            Assert.assertNotEquals(providerMetaData.getId(), "federated-repo");
        }
    }

    @DataProvider
    public Object[][] addFederationProvider() {
        return new Object[][] {
                {true},
                {false}
        };
    }
}
